# -*- coding: utf-8 -*-

'''
Coverity plugin

Sphinx extension for restructured text that adds Coverity reporting to documentation.
See README.rst for more details.
'''
from __future__ import print_function

from hashlib import sha256
from os import environ, mkdir, path
from re import findall

import pkg_resources
from docutils import nodes
from docutils.parsers.rst import Directive, directives
from sphinx import __version__ as sphinx_version
from sphinx.environment import NoUri
from urlextract import URLExtract

from mlx.coverity_services import CoverityConfigurationService, CoverityDefectService
import matplotlib as mpl
if not environ.get('DISPLAY'):
    mpl.use('Agg')
import matplotlib.pyplot as plt  # noqa: E731

try:
    # For Python 3.0 and later
    from urllib.error import URLError, HTTPError
except ImportError:
    # Fall back to Python 2's urllib2
    from urllib2 import URLError, HTTPError
if sphinx_version >= '1.6.0':
    from sphinx.util.logging import getLogger


def report_warning(env, msg, docname, lineno=None):
    '''Convenience function for logging a warning

    Args:
        msg (str): Message of the warning
        docname (str): Name of the document on which the error occured
        lineno (str): Line number in the document on which the error occured
    '''
    if sphinx_version >= '1.6.0':
        logger = getLogger(__name__)
        if lineno is not None:
            logger.warning(msg, location=(docname, lineno))
        else:
            logger.warning(msg, location=docname)
    else:
        env.warn(docname, msg, lineno=lineno)


def report_info(env, msg, nonl=False):
    '''Convenience function for information printing

    Args:
        msg (str): Message of the warning
        nonl (bool): True when no new line at end
    '''
    if sphinx_version >= '1.6.0':
        logger = getLogger(__name__)
        logger.info(msg, nonl=nonl)
    else:
        env.info(msg, nonl=nonl)


def pct_wrapper(sizes):
    """ Helper function for matplotlib which returns the percentage and the absolute size of the slice.

    Args:
        sizes (list): List containing the amount of elements per slice.
    """
    def make_pct(pct):
        absolute = int(round(pct / 100 * sum(sizes)))
        return "{:.0f}%\n({:d})".format(pct, absolute)
    return make_pct


# -----------------------------------------------------------------------------
# Declare new node types (based on others):
class CoverityDefect(nodes.General, nodes.Element):
    '''Coverity defect'''
    pass


# -----------------------------------------------------------------------------
# Directives
class CoverityDefectListDirective(Directive):
    """
    Directive to generate a list of defects.

    Syntax::

      .. coverity-list:: title
         :col: list of columns to be displayed
         :widths: list of predefined column widths
         :chart: display chart that labels each allowed <<attribute>> value
         :checker: filter for only these checkers
         :impact: filter for only these impacts
         :kind: filter for only these kinds
         :classification: filter for only these classifications
         :action: filter for only these actions
         :component: filter for only these components
         :cwe: filter for only these CWE rating
         :cid: filter only these cid
    """
    # Optional argument: title (whitespace allowed)
    optional_arguments = 1
    final_argument_whitespace = True
    # Options
    option_spec = {'class': directives.class_option,
                   'col': directives.unchanged,
                   'widths': directives.value_or(('auto', 'grid'),
                                                 directives.positive_int_list),
                   'chart': directives.unchanged,
                   'checker': directives.unchanged,
                   'impact': directives.unchanged,
                   'kind': directives.unchanged,
                   'classification': directives.unchanged,
                   'action': directives.unchanged,
                   'component': directives.unchanged,
                   'cwe': directives.unchanged,
                   'cid': directives.unchanged,
                   }
    # Content disallowed
    has_content = False

    def run(self):
        item_list_node = CoverityDefect('')

        # Process title (optional argument)
        if len(self.arguments) > 0:
            item_list_node['title'] = self.arguments[0]
        else:
            item_list_node['title'] = 'Coverity report'

        # Process ``col`` option
        if 'col' in self.options:
            item_list_node['col'] = self.options['col'].split(',')
        elif 'chart' not in self.options:
            item_list_node['col'] = 'CID,Classification,Action,Comment'.split(',')  # use default colums
        else:
            item_list_node['col'] = []  # don't display a table if the ``chart`` option is present without ``col``

        # Process ``widths`` option
        if 'widths' in self.options:
            item_list_node['widths'] = self.options['widths']
        else:
            item_list_node['widths'] = ''

        # Process ``chart`` option
        if 'chart' in self.options:
            if ':' in self.options['chart']:
                item_list_node['chart_attribute'] = self.options['chart'].split(':')[0].capitalize()
            else:
                item_list_node['chart_attribute'] = 'Classification'

            parameters = self.options['chart'].split(':')[-1]  # str
            item_list_node['chart'] = parameters.split(',')  # list
            # try to convert parameters to int, in case a min slice size is defined instead of filter options
            try:
                item_list_node['min_slice_size'] = int(item_list_node['chart'][0])
                item_list_node['chart'] = []  # only when a min slice size is defined
            except ValueError:
                item_list_node['min_slice_size'] = 1
        else:
            item_list_node['chart'] = ''

        # Process the optional filters
        filters = ['checker', 'impact', 'kind', 'classification', 'action', 'component', 'cwe', 'cid']
        for fil in filters:
            if fil in self.options:
                item_list_node[fil] = self.options[fil]
            else:
                item_list_node[fil] = None

        return [item_list_node]


class SphinxCoverityConnector():
    def __init__(self):
        """
        Initialize the object by setting error variable to false
        """
        self.coverity_login_error = False

    def initialize_environment(self, app):
        """
        Perform initializations needed before the build process starts.
        """

        # LaTeX-support: since we generate empty tags, we need to relax the verbosity of that error
        if 'preamble' not in app.config.latex_elements:
            app.config.latex_elements['preamble'] = ''
        app.config.latex_elements['preamble'] += '''\
    \\makeatletter
    \\let\@noitemerr\\relax
    \\makeatother'''

        env = app.builder.env

        # Login to Coverity and obtain stream information
        try:
            report_info(env, 'Login to Coverity server... ', True)
            coverity_conf_service = CoverityConfigurationService(app.config.coverity_credentials['transport'],
                                                                 app.config.coverity_credentials['hostname'],
                                                                 app.config.coverity_credentials['port'])
            coverity_conf_service.login(app.config.coverity_credentials['username'],
                                        app.config.coverity_credentials['password'])
            report_info(env, 'done')

            report_info(env, 'obtaining stream information... ', True)
            stream = coverity_conf_service.get_stream(app.config.coverity_credentials['stream'])
            if stream is None:
                raise ValueError('No such Coverity stream [%s] found on [%s]' %
                                 (app.config.coverity_credentials['stream'], coverity_conf_service.get_service_url()))
            report_info(env, 'done')

            # Get Stream's project name
            report_info(env, 'obtaining project name from stream... ', True)
            self.project_name = coverity_conf_service.get_project_name(stream)
            report_info(env, 'done')
            self.coverity_service = CoverityDefectService(coverity_conf_service)
            self.coverity_service.login(app.config.coverity_credentials['username'],
                                        app.config.coverity_credentials['password'])
        except (URLError, HTTPError, Exception, ValueError) as error_info:
            self.coverity_login_error_msg = error_info
            report_info(env, 'failed with: %s' % error_info)
            self.coverity_login_error = True

    # -----------------------------------------------------------------------------
    # Event handlers
    def process_coverity_nodes(self, app, doctree, fromdocname):
        """
        This function should be triggered upon ``doctree-resolved event``

        Obtain information from Coverity server and generate a table.
        """
        env = app.builder.env

        if self.coverity_login_error:
            # Create failed topnode
            for node in doctree.traverse(CoverityDefect):
                top_node = create_top_node("Failed to connect to Coverity Server")
                node.replace_self(top_node)
            report_warning(env, 'Connection failed: %s' % self.coverity_login_error_msg, fromdocname)
            return

        # Item matrix:
        # Create table with related items, printing their target references.
        # Only source and target items matching respective regexp shall be included
        for node in doctree.traverse(CoverityDefect):
            top_node = create_top_node(node['title'])

            # Initialize table
            if node['col']:
                table = nodes.table()
                table.set_class('longtable')
                if node['widths'] == 'auto':
                    table['classes'] += ['colwidths-auto']
                elif node['widths']:  # "grid" or list of integers
                    table['classes'] += ['colwidths-given']
                tgroup = nodes.tgroup()

                for _ in node['col']:
                    tgroup += [nodes.colspec(colwidth=5)]

                tgroup += nodes.thead('', create_row(node['col']))

                if isinstance(node['widths'], list):
                    colspecs = [child for child in tgroup.children
                                if child.tagname == 'colspec']
                    for colspec, col_width in zip(colspecs, node['widths']):
                        colspec['colwidth'] = col_width

                tbody = nodes.tbody()

                tgroup += tbody
                table += tgroup

            # Get items from server
            report_info(env, 'obtaining defects... ', True)
            try:
                defects = self.coverity_service.get_defects(self.project_name, app.config.coverity_credentials['stream'],   # noqa: E501
                                                            checker=node['checker'], impact=node['impact'], kind=node['kind'],  # noqa: E501
                                                            classification=node['classification'], action=node['action'],   # noqa: E501
                                                            component=node['component'], cwe=node['cwe'], cid=node['cid'])  # noqa: E501
            except (URLError, AttributeError, Exception) as err:
                report_warning(env, 'failed with %s' % err, fromdocname)
                continue
            report_info(env, "%d received" % (defects['totalNumberOfRecords']))
            report_info(env, "building defects table and/or chart... ", True)

            # Initialize dictionary to store counters
            if isinstance(node['chart'], list):
                chart_labels = {}
                combined_labels = {}
                for label in node['chart']:
                    if '+' in label:
                        combined_labels[label] = label.split('+')
                    for attr_val in label.split('+'):
                        if attr_val in chart_labels.keys():
                            report_warning(env,
                                           "Attribute value '%s' should not be specified more than once in chart "
                                           "option." % attr_val,
                                           fromdocname)
                        chart_labels[attr_val] = 0

            column_map = {
                'Cid': 'cid',
                'Category': 'displayCategory',
                'Impact': 'displayImpact',
                'Issue': 'displayIssueKind',
                'Type': 'displayType',
                'Checker': 'checkerName',
                'Component': 'componentName',
            }
            try:
                for defect in defects['mergedDefects']:
                    if node['col']:
                        row = nodes.row()

                        # go through each col and decide if it is there or we print empty
                        for item_col in node['col']:
                            if 'CID' == item_col:
                                # CID is default and even if it is in disregard
                                row += create_cell(str(defect['cid']),
                                                   url=self.coverity_service.get_defect_url(
                                                       app.config.coverity_credentials['stream'],
                                                       str(defect['cid'])))
                            elif 'Category' == item_col:
                                row += create_cell(defect['displayCategory'])
                            elif 'Impact' == item_col:
                                row += create_cell(defect['displayImpact'])
                            elif 'Issue' == item_col:
                                row += create_cell(defect['displayIssueKind'])
                            elif 'Type' == item_col:
                                row += create_cell(defect['displayType'])
                            elif 'Checker' == item_col:
                                row += create_cell(defect['checkerName'])
                            elif 'Component' == item_col:
                                row += create_cell(defect['componentName'])
                            elif 'Comment' == item_col:
                                text = str(cov_attribute_value_to_col(defect, 'Comment').children[0].children[0])
                                contents = create_paragraph_with_links(text, str(defect['cid']), app, fromdocname)
                                row += nodes.entry('', contents)
                            elif 'Classification' == item_col:
                                row += cov_attribute_value_to_col(defect, 'Classification')
                            elif 'Action' == item_col:
                                row += cov_attribute_value_to_col(defect, 'Action')
                            elif 'Status' == item_col:
                                row += cov_attribute_value_to_col(defect, 'DefectStatus')
                            else:
                                # generic check which, if it is missing, prints empty cell anyway
                                row += cov_attribute_value_to_col(defect, item_col)
                        tbody += row

                    if isinstance(node['chart'], list):
                        if node['chart_attribute'] in column_map.keys():
                            classification_value = str(defect[column_map[node['chart_attribute']]])
                        else:
                            col = cov_attribute_value_to_col(defect, node['chart_attribute'])
                            classification_value = str(col.children[0].children[0])  # get text in paragraph of column

                        if classification_value in chart_labels.keys():
                            chart_labels[classification_value] += 1
                        elif not node['chart']:  # remove those that don't comply with min_slice_length
                            chart_labels[classification_value] = 1

            except AttributeError as err:
                report_info(env, 'No issues matching your query or empty stream. %s' % err)
                top_node += nodes.paragraph(text='No issues matching your query or empty stream')

            if node['col']:
                top_node += table

            if isinstance(node['chart'], list):
                for new_label, old_labels in combined_labels.items():
                    count = 0
                    for old_label in old_labels:
                        count += chart_labels.pop(old_label)  # remove old_label and store its count
                    chart_labels[new_label] = count  # add combined count under new_label

                # only keep those labels that comply with the min_slice_size requirement
                chart_labels = {label: count for label, count in chart_labels.items()
                                if count >= node['min_slice_size']}

                total_labeled = sum(list(chart_labels.values()))
                other_count = defects['totalNumberOfRecords'] - total_labeled
                if other_count:
                    chart_labels['Other'] = other_count

                labels = list(chart_labels.keys())
                sizes = list(chart_labels.values())
                fig, axes = plt.subplots()
                fig.set_size_inches(10, 6)
                axes.pie(sizes, labels=labels, autopct=pct_wrapper(sizes), startangle=90)
                axes.axis('equal')
                folder_name = path.join(env.app.srcdir, '_images')
                if not path.exists(folder_name):
                    mkdir(folder_name)
                hash_string = ''
                for pie_slice in axes.__dict__['texts']:
                    hash_string += str(pie_slice)
                hash_value = sha256(hash_string.encode()).hexdigest()  # create hash value based on chart parameters
                rel_file_path = path.join('_images', 'piechart-{}.png'.format(hash_value))
                if rel_file_path not in env.images.keys():
                    fig.savefig(path.join(env.app.srcdir, rel_file_path), format='png')
                    # store file name in build env
                    env.images[rel_file_path] = ['_images', path.split(rel_file_path)[-1]]

                image_node = nodes.image()
                image_node['uri'] = rel_file_path
                image_node['candidates'] = '*'  # look at uri value for source path, relative to the srcdir folder
                top_node += image_node

            report_info(env, "done")
            node.replace_self(top_node)
    #        try:
    #            percentage = int(100 * count_covered / count_total)
    #        except ZeroDivisionError:
    #            percentage = 0
    #        disp = 'Statistics: {cover} out of {total} covered: {pct}%'.format(cover=count_covered,
    #                                                                           total=count_total,
    #                                                                           pct=percentage)
    #        if node['graph']:
    #            p_node = nodes.paragraph()
    #            txt = nodes.Text(disp)
    #            p_node += txt
    #            top_node += p_node
    #
    #        top_node += table
    #


def create_ref_node(contents, url):
    p_node = nodes.paragraph()
    itemlink = nodes.reference()
    itemlink['refuri'] = url
    itemlink.append(nodes.Text(contents))
    targetid = nodes.make_id(contents)
    target = nodes.target('', '', ids=[targetid])
    p_node += target
    p_node += itemlink
    return p_node


def create_top_node(title):
    top_node = nodes.container()
    admon_node = nodes.admonition()
    title_node = nodes.title()
    title_node += nodes.Text(title)
    admon_node += title_node
    top_node += admon_node
    return top_node


def create_cell(contents, url=None):
    if isinstance(contents, str):
        if url is not None:
            contents = create_ref_node(contents, url)
        else:
            contents = nodes.paragraph(text=contents)

    return nodes.entry('', contents)


def create_row(cells):
    return nodes.row('', *[create_cell(c) for c in cells])


def cov_attribute_value_to_col(defect, name):
    """
    Search defects array and return value for name
    """
    col = create_cell(" ")

    for attribute in defect['defectStateAttributeValues']:
        if attribute['attributeDefinitionId'][0] == name:
            try:
                col = create_cell(attribute['attributeValueId'][0])
            except (AttributeError, IndexError):
                col = create_cell(" ")
    return col


def create_paragraph_with_links(text, *args):
    """
    Create a paragraph with the provided text. Hyperlinks are made interactive, and traceability item IDs get linked to
    their definition.
    """
    contents = nodes.paragraph()
    remaining_text = text
    link_to_urls(contents, remaining_text, *args)
    return contents


def link_to_urls(contents, text, *args):
    """
    Makes URLs interactive and passes other text to link_to_item_ids, which treats the item IDs.
    """
    remaining_text = text
    extractor = URLExtract()
    urls = extractor.find_urls(remaining_text)
    for url in urls:
        text_before = remaining_text.split(url)[0]
        if text_before:
            link_to_item_ids(contents, text_before, *args)

        ref_node = nodes.reference()
        ref_node['refuri'] = url
        ref_node.append(nodes.Text(url))
        contents += ref_node

        remaining_text = remaining_text.replace(text_before + url, '', 1)

    if remaining_text:
        link_to_item_ids(contents, text, *args)


def link_to_item_ids(contents, text, cid, app, docname):
    """
    Makes a link of item IDs when they are found in a traceability collection and adds all other text to the paragraph.
    """
    remaining_text = text
    item_matches = findall(app.config.TRACEABILITY_ITEM_ID_REGEX, remaining_text)
    for item in item_matches:
        text_before = remaining_text.split(item)[0]
        if text_before:
            contents.append(nodes.Text(text_before))
        ref_node = make_internal_item_ref(app, docname, item, cid)
        if ref_node is None:  # no link could be made
            ref_node = nodes.Text(item)
        contents.append(ref_node)

        remaining_text = remaining_text.replace(text_before + item, '', 1)

    if remaining_text:
        contents.append(nodes.Text(remaining_text))  # no URL or item ID in this text


def make_internal_item_ref(app, fromdocname, item, cid):
    """
    Creates and returns a reference node for an item or returns None when the item cannot be found in the traceability
    collection. A warning is raised when a traceability collection exists, but an item ID cannot be found in it.
    """
    env = app.builder.env

    if not hasattr(env, 'traceability_collection'):
        return None

    item_info = env.traceability_collection.get_item(item)
    if not item_info:
        report_warning(env,
                       "CID %s: Could not find item ID '%s' in traceability collection." % (cid, item),
                       fromdocname)
        return None

    ref_node = nodes.reference('', '')
    ref_node['refdocname'] = item_info.docname
    try:
        ref_node['refuri'] = app.builder.get_relative_uri(fromdocname, item_info.docname) + '#' + item
    except NoUri:
        return None
    ref_node.append(nodes.Text(item))
    return ref_node


# Extension setup
def setup(app):
    '''Extension setup'''

    # Create default configuration. Can be customized in conf.py
    app.add_config_value('coverity_credentials',
                         {
                             'hostname': 'scan.coverity.com',
                             'port': '8080',
                             'transport': 'http',
                             'username': 'reporter',
                             'password': 'coverity',
                             'stream': 'some_coverty_stream',
                         },
                         'env')
    app.add_config_value('TRACEABILITY_ITEM_ID_REGEX', r"([A-Z_]+-[A-Z0-9_]+)", 'env')

    app.add_node(CoverityDefect)

    sphinx_coverity_connector = SphinxCoverityConnector()

    app.add_directive('coverity-list', CoverityDefectListDirective)

    app.connect('doctree-resolved', sphinx_coverity_connector.process_coverity_nodes)

    app.connect('builder-inited', sphinx_coverity_connector.initialize_environment)

    try:
        return {'version': '%(prog)s {version}'.format(version=pkg_resources.require('mlx.coverity')[0].version)}
    except LookupError:
        return {'version': 'dev'}



#!/usr/bin/python

'''Services and other utilities for Coverity scripting'''

# General
import csv
import logging
import re
try:
    # For Python 3.0 and later
    from urllib.error import URLError
except ImportError:
    # Fall back to Python 2's urllib2
    from urllib2 import URLError

# For Coverity - SOAP
from suds.client import Client
from suds.wsse import Security, UsernameToken

# -- Default values -- and global settings

DEFAULT_WS_VERSION = 'v9'

# Coverity built in Impact statuses
IMPACT_LIST = {'High', 'Medium', 'Low'}

KIND_LIST = {'QUALITY', 'SECURITY', 'TEST'}

# Coverity built in Classifications
CLASSIFICATION_LIST = {'Unclassified', 'Pending', 'False Positive', 'Intentional', 'Bug', 'Untested', 'No Test Needed'}

# Coverity built in Actions
ACTION_LIST = {'Undecided', 'Fix Required', 'Fix Submitted', 'Modeling Required', 'Ignore', 'On hold',
               'For Interest Only'}

ISSUE_KIND_2_LABEL = {'QUALITY': 'Quality', 'SECURITY': 'Security', 'Various': 'Quality/Security', 'TEST': 'Testing'}


# names of Coverity Triage/Attribute fields
EXT_REFERENCE_ATTR_NAME = "Ext. Reference"
DEFECT_STATUS_ATTR_NAME = "DefectStatus"
CLASSIFICATION_ATTR_NAME = "Classification"
ACTION_ATTR_NAME = "Action"
COMMENT_ATTR_NAME = "Comment"


def parse_two_part_term(term, delim=','):
    '''Parse a term assuming [ [part1],[part2] ]'''
    valid = False
    part1 = ""
    part2 = ""
    if term.find(delim) != -1:
        valid = True
        field1 = term.split(delim, 1)[0]
        if bool(field1):
            part1 = field1
        field2 = term.rsplit(delim, 1)[-1]
        if bool(field2):
            part2 = field2
    return valid, part1, part2


def compare_strings(str_a, str_b):
    '''Compare strings for equivalence

    some leniency allowed such as spaces and casing
    '''
    if re.match(str_b, str_a, flags=re.IGNORECASE):
        return True
    # ignore embedded spaces and some odd punctuation characters ("todo" = "To-Do")
    str_a2 = re.sub(r'[.:\-_ ]', '', str_a)
    str_b2 = re.sub(r'[:\-_ ]', '', str_b)  # don't remove dot (part of regex?)
    if re.match(str_b2, str_a2, flags=re.IGNORECASE):
        return True
    return False


class Service(object):
    '''
    Basic endpoint Service
    '''

    def __init__(self, transport, hostname, port, ws_version=DEFAULT_WS_VERSION):
        self.set_transport(transport)
        self.set_hostname(hostname)
        self.set_port(port)
        self.set_ws_version(ws_version)
        self.client = None

    def set_transport(self, transport):
        '''Set transport protocol'''
        self.transport = transport

    def get_transport(self):
        '''Get transport protocol'''
        return self.transport

    def set_hostname(self, hostname):
        '''Set hostname for service'''
        self.hostname = hostname

    def get_hostname(self):
        '''Get hostname for service'''
        return self.hostname

    def set_port(self, port):
        '''Set port for service'''
        self.port = port

    def get_port(self):
        '''Get port for service'''
        return self.port

    def set_ws_version(self, ws_version):
        '''Set WS version for service'''
        self.ws_version = ws_version

    def get_ws_version(self):
        '''Get WS version for service'''
        return self.ws_version

    def get_service_url(self, path=''):
        '''Get Service url with given path'''
        url = self.transport + '://' + self.hostname
        if bool(self.port):
            url += ':' + self.port
        if path:
            url += path
        return url

    def get_ws_url(self, service):
        '''Get WS url with given service'''
        return self.get_service_url('/ws/' + self.ws_version + '/' + service + '?wsdl')

    def login(self, username, password):
        '''Login to Coverity using given username and password'''
        security = Security()
        token = UsernameToken(username, password)
        security.tokens.append(token)
        self.client.set_options(wsse=security)


class CoverityConfigurationService(Service):
    '''
    Coverity Configuration Service (WebServices)
    '''

    def __init__(self, transport, hostname, port, ws_version=DEFAULT_WS_VERSION):
        super(CoverityConfigurationService, self).__init__(transport, hostname, port, ws_version)
        self.checkers = None
        url = self.get_ws_url('configurationservice')
        logging.getLogger('suds.client').setLevel(logging.CRITICAL)
        try:
            self.client = Client(url)
            logging.info("Validated presence of Coverity Configuration Service [%s]", url)
        except URLError:
            self.client = None
            logging.critical("No such Coverity Configuration Service [%s]", url)
            raise

    def login(self, username, password):
        '''Login to Coverity Configuration service using given username and password'''
        super(CoverityConfigurationService, self).login(username, password)
        version = self.get_version()
        if version is None:
            raise RuntimeError("Authentication to [%s] FAILED for [%s] account - check password"
                               % (self.get_service_url(), username))
        else:
            logging.info("Authentication to [%s] using [%s] account was OK - version [%s]",
                         self.get_service_url(), username, version.externalVersion)

    def get_version(self):
        '''Get the version of the service, can be used as a means to validate access permissions'''
        try:
            return self.client.service.getVersion()
        except URLError:
            return None

    @staticmethod
    def get_project_name(stream):
        '''Get the project name from the stream object'''
        return stream.primaryProjectId.name

    @staticmethod
    def get_triage_store(stream):
        '''Get the name of the triaging store from the stream object'''
        return stream.triageStoreId.name

    def get_stream(self, stream_name):
        '''Get the stream object from the stream name'''
        filter_spec = self.client.factory.create('streamFilterSpecDataObj')

        # use stream name as an initial glob pattern
        filter_spec.namePattern = stream_name

        # get all the streams that match
        streams = self.client.service.getStreams(filter_spec)

        # find the one with a matching name
        for stream in streams:
            if compare_strings(stream.id.name, stream_name):
                return stream
        return None

    # get a list of the snapshots in a named stream
    def get_snapshot_for_stream(self, stream_name):
        '''Get snapshot object for given stream name'''
        stream_id = self.client.factory.create('streamIdDataObj')
        stream_id.name = stream_name
        # optional filter specification
        filter_spec = self.client.factory.create('snapshotFilterSpecDataObj')
        # return a list of snapshotDataObj
        return self.client.service.getSnapshotsForStream(stream_id, filter_spec)

    @staticmethod
    def get_snapshot_id(snapshots, idx=1):
        '''Get the nth snapshot (base 1) - minus numbers to count from the end backwards (-1 = last)'''
        if bool(idx):
            num_snapshots = len(snapshots)
            if idx < 0:
                required = num_snapshots + idx + 1
            else:
                required = idx

            if abs(required) > 0 and abs(required) <= num_snapshots:
                # base zero
                return snapshots[required - 1].id
        return 0

    def get_snapshot_detail(self, snapshot_id):
        '''Get detailed information about a single snapshot'''
        snapshot = self.client.factory.create('snapshotIdDataObj')
        snapshot.id = snapshot_id
        # return a snapshotInfoDataObj
        return self.client.service.getSnapshotInformation(snapshot)

    def get_checkers(self):
        '''Get a list of checkers from the service'''
        if not self.checkers:
            self.checkers = self.client.service.getCheckerNames()
        return self.checkers

    @staticmethod
    def add_filter_rqt(name, req_csv, valid_list, filter_list, allow_regex=False):
        '''Lookup the list of given filter possibility, add to filter spec and return a validated list'''
        logging.info('Validate required %s [%s]', name, req_csv)
        validated = ""
        delim = ""
        for field in req_csv.split(','):
            if not valid_list or field in valid_list:
                logging.info('Classification [%s] is valid', field)
                filter_list.append(field)
                validated += delim + field
                delim = ","
            elif allow_regex:
                pattern = re.compile(field)
                for element in valid_list:
                    if pattern.search(element) and element not in filter_list:
                        filter_list.append(element)
                        validated += delim + element
                        delim = ","
            else:
                logging.error('Invalid %s filter: %s', name, field)
        return validated


class CoverityDefectService(Service):
    '''
    Coverity Defect Service (WebServices)
    '''

    def __init__(self, config_service):
        '''Create a Defect Service, bound to the given Configuration Service'''
        super(CoverityDefectService, self).__init__(config_service.get_transport(),
                                                    config_service.get_hostname(),
                                                    config_service.get_port(),
                                                    config_service.get_ws_version())
        self.config_service = config_service
        self.filters = ""
        # logging.getLogger('suds.client').setLevel(logging.DEBUG)
        url = self.get_ws_url('defectservice')
        try:
            self.client = Client(url)
            logging.info("Validated presence of Coverity Defect Service [%s]", url)
        except URLError:
            self.client = None
            logging.critical("No such Coverity Defect Service [%s]", url)
            raise

    def get_defects(self, project, stream, checker=None, impact=None, kind=None, classification=None, action=None,
                    component=None, cwe=None, cid=None, custom=None):
        '''
        Get a list of defects for given stream, with some query criteria

        Arguments:
        project (str) - Name of the project to query
        stream (str) - Name of the stream to query
        checker (str) - A CSV list of the checker(s) to query
        impact (str) - A CSV list of the impact(s) to query
        kind (str) - A CSV list of the kind(s) to query
        classification (str) - A CSV list of the classification(s) to query
        action (str) - A CSV list of the action(s) to query
        component (str) - A CSV list of component(s) to include/exclude in the query
        exclude_component(bool) - True for excluding the given component(s), false for including them
        cwe (str) - A CSV list of the cwe(s) to query
        cid (str) - A CSV list of the cid(s) to query
        custom (str) - A custom query
        '''
        logging.info('Querying Coverity for defects in project [%s] stream [%s] ...', project, stream)

        # define the project
        project_id = self.client.factory.create('projectIdDataObj')
        project_id.name = project

        # and the stream
        stream_id = self.client.factory.create('streamIdDataObj')
        stream_id.name = stream

        # create filter spec
        filter_spec = self.client.factory.create('snapshotScopeDefectFilterSpecDataObj')

        # only for this stream
        filter_spec.streamIncludeNameList.append(stream_id)

        # apply any filter on checker names
        if bool(checker):
            logging.info('Using checker filter [%s]', checker)
            self.config_service.get_checkers()
            validated = self.config_service.add_filter_rqt('Checker', checker, self.config_service.checkers,
                                                           filter_spec.checkerList, allow_regex=True)
            logging.info('Resolves to [%s]', validated)
            if validated:
                self.filters += ("<Checker(%s)> " % validated)

        # apply any filter on impact status
        if bool(impact):
            logging.info('Using impact status filter [%s]', impact)
            validated = self.config_service.add_filter_rqt('Impact', impact, IMPACT_LIST, filter_spec.impactNameList)
            logging.info('Resolves to [%s]', validated)
            if validated:
                self.filters += ("<Impact(%s)> " % validated)

        # apply any filter on issue kind
        if bool(kind):
            logging.info('Using issue kind filter [%s]', kind)
            validated = self.config_service.add_filter_rqt('Kind', kind, KIND_LIST, filter_spec.issueKindList)
            logging.info('Resolves to [%s]', validated)
            if validated:
                self.filters += ("<Kind(%s)> " % validated)

        # apply any filter on classification
        if bool(classification):
            logging.info('Using classification filter [%s]', classification)
            validated = self.config_service.add_filter_rqt('Classification', classification, CLASSIFICATION_LIST,
                                                           filter_spec.classificationNameList)
            logging.info('Resolves to [%s]', validated)
            if validated:
                self.filters += ("<Classification(%s)> " % validated)

        # apply any filter on action
        if bool(action):
            logging.info('Using action filter [%s]', action)
            validated = self.config_service.add_filter_rqt('Action', action, ACTION_LIST, filter_spec.actionNameList)
            logging.info('Resolves to [%s]', validated)
            if validated:
                self.filters += ("<Action(%s)> " % validated)

        # apply any filter on Components
        if bool(component):
            logging.info('Using Component filter [%s]', component)
            parser = csv.reader([component])

            for fields in parser:
                for _, field in enumerate(fields):
                    field = field.strip()
                    component_id = self.client.factory.create('componentIdDataObj')
                    component_id.name = field
                    filter_spec.componentIdList.append(component_id)
            self.filters += ("<Components(%s)> " % (component))

        # apply any filter on CWE values
        if bool(cwe):
            logging.info('Using CWE filter [%s]', cwe)
            validated = self.config_service.add_filter_rqt('CWE', cwe, None, filter_spec.cweList)
            logging.info('Resolves to [%s]', validated)
            if validated:
                self.filters += ("<CWE(%s)> " % validated)

        # apply any filter on CID values
        if bool(cid):
            logging.info('Using CID filter [%s]', cid)
            validated = self.config_service.add_filter_rqt('CID', cid, None, filter_spec.cidList)
            logging.info('Resolves to [%s]', validated)
            if validated:
                self.filters += ("<CID(%s)> " % validated)

        # if a special custom attribute value requirement
        if bool(custom):
            logging.info('Using attribute filter [%s]', custom)
            # split the name:value[;name:value1[,value2]]
            for fields in csv.reader([custom], delimiter=';'):
                for i, name_value_pair in enumerate(fields):
                    name_value_pair = name_value_pair.strip()
                    valid, name, values = parse_two_part_term(name_value_pair, ':')
                    if valid:
                        logging.info("attr (%d) [%s] = any of ...", i + 1, name)

                        attribute_definition_id = self.client.factory.create('attributeDefinitionIdDataObj')
                        attribute_definition_id.name = name

                        filter_map = self.client.factory.create('attributeDefinitionValueFilterMapDataObj')
                        filter_map.attributeDefinitionId = attribute_definition_id

                        # if there are multiple values delimited with comma
                        for value_fields in csv.reader([values], delimiter=','):
                            for _, value in enumerate(value_fields):
                                logging.info("             [%s]", value)

                                attribute_value_id = self.client.factory.create('attributeValueIdDataObj')
                                attribute_value_id.name = value

                                filter_map.attributeValueIds.append(attribute_value_id)

                        filter_spec.attributeDefinitionValueFilterMap.append(filter_map)
                    else:
                        raise ValueError('Invalid custom attribute definition [%s]', name_value_pair)
            self.filters += ("<Attrs(%s)> " % custom)

        # create page spec
        page_spec = self.client.factory.create('pageSpecDataObj')
        page_spec.pageSize = 9999
        page_spec.sortAscending = True
        page_spec.startIndex = 0

        # create snapshot scope
        snapshot_scope = self.client.factory.create('snapshotScopeSpecDataObj')

        snapshot_scope.showOutdatedStreams = False
        snapshot_scope.compareOutdatedStreams = False

        snapshot_scope.showSelector = 'last()'
        snapshot_scope.compareSelector = 'last()'

        logging.info('Running Coverity query...')
        return self.client.service.getMergedDefectsForSnapshotScope(project_id, filter_spec,
                                                                    page_spec, snapshot_scope)

    def get_defect(self, cid, stream):
        '''Get the details pertaining a specific CID - it may not have defect instance details if newly eliminated
        (fixed)'''
        logging.info('Fetching data for CID [%s] in stream [%s] ...', cid, stream)

        merged_defect_id = self.client.factory.create('mergedDefectIdDataObj')
        merged_defect_id.cid = cid

        filter_spec = self.client.factory.create('streamDefectFilterSpecDataObj')
        filter_spec.includeDefectInstances = True
        filter_spec.includeHistory = True

        stream_id = self.client.factory.create('streamIdDataObj')
        stream_id.name = stream
        filter_spec.streamIdList.append(stream_id)

        return self.client.service.getStreamDefects(merged_defect_id, filter_spec)

    def add_attribute_name_and_value(self, defect_state_spec, attr_name, attr_value):
        '''Add attribute name and value to given defect state specification'''

        # name value pair to update
        attribute_definition_id = self.client.factory.create('attributeDefinitionIdDataObj')
        attribute_definition_id.name = attr_name

        attribute_value_id = self.client.factory.create('attributeValueIdDataObj')
        attribute_value_id.name = attr_value

        # wrap the name/value pair
        defect_state_attr_value = self.client.factory.create('defectStateAttributeValueDataObj')
        defect_state_attr_value.attributeDefinitionId = attribute_definition_id
        defect_state_attr_value.attributeValueId = attribute_value_id

        # add to our list
        defect_state_spec.defectStateAttributeValues.append(defect_state_attr_value)

    # update the external reference id to a third party
    def update_ext_reference_attribute(self, cid, triage_store, ext_ref_id, ccomment=None):
        '''Update external reference attribute for given CID'''
        logging.info('Updating Coverity - CID [%s] in TS [%s] with Ext Ref [%s]', cid, triage_store, ext_ref_id)

        # triage store identifier
        triage_store_id = self.client.factory.create('triageStoreIdDataObj')
        triage_store_id.name = triage_store

        # CID to update
        merged_defect_id = self.client.factory.create('mergedDefectIdDataObj')
        merged_defect_id.cid = cid

        # if an ext ref id value supplied
        if bool(ext_ref_id):
            attr_value = ext_ref_id
            comment_value = 'Automatically recorded reference to new JIRA ticket.'
        else:
            # set to a space - which works as a blank without the WS complaining :-)
            attr_value = " "
            comment_value = 'Automatically cleared former JIRA ticket reference.'

        # if a Coverity comment to tag on the end
        if bool(ccomment):
            comment_value += " " + ccomment
        logging.info('Comment = [%s]', comment_value)

        defect_state_spec = self.client.factory.create('defectStateSpecDataObj')

        # name value pairs to add to this update
        self.add_attribute_name_and_value(defect_state_spec, EXT_REFERENCE_ATTR_NAME, attr_value)
        self.add_attribute_name_and_value(defect_state_spec, COMMENT_ATTR_NAME, comment_value)

        # apply the update
        return self.client.service.updateTriageForCIDsInTriageStore(triage_store_id, merged_defect_id,
                                                                    defect_state_spec)

    @staticmethod
    def get_instance_impact(stream_defect, instance_number=1):
        '''Get the current impact of the 'nth' incident of this issue (High/Medium/Low)'''
        counter = instance_number
        for instance in stream_defect.defectInstances:
            counter -= 1
            if counter == 0:
                return instance.impact.name
        return ""

    @staticmethod
    def get_value_for_named_attribute(stream_defect, attr_name):
        '''Lookup the value of a named attribute'''
        logging.info('Get value for cov attribute [%s]', attr_name)
        for attr_value in stream_defect.defectStateAttributeValues:
            if compare_strings(attr_value.attributeDefinitionId.name, attr_name):
                logging.info('Resolves to [%s]', attr_value.attributeValueId.name)
                return str(attr_value.attributeValueId.name)
        logging.warning('Value for attribute [%s] not found', attr_name)
        return ""

    @staticmethod
    def get_event_attribute_value(defect_state, name, value=None):
        '''Get specified attribute was set to given matching value'''
        if bool(value):
            logging.info('Searching for attribute [%s] with value [%s]', name, value)
        else:
            logging.info('Searching for attribute [%s]', name)

        for attr_value in defect_state.defectStateAttributeValues:
            # check if we have the named attribute
            if compare_strings(attr_value.attributeDefinitionId.name, name):
                # if any value supplied or it matches requirement
                if bool(attr_value.attributeValueId.name) and\
                   (not value or compare_strings(attr_value.attributeValueId.name, value)):
                    logging.info('Found [%s] = [%s]',
                                 attr_value.attributeDefinitionId.name, attr_value.attributeValueId.name)
                    return True, attr_value.attributeValueId.name
                else:
                    # break attribute name search - either no value or it doesn't match
                    break
        logging.warning('Event for attribute [%s] not found', name)
        return False, None

    def seek_nth_match(self, event_history, nth_event, attr_name, attr_value):
        '''Seek for a given attribute name-value pair in the triaging history'''
        num_match = 0
        for defect_state in event_history:
            # look for the attribute name-value pair in this triage event
            req_event_found, req_attr_value = self.get_event_attribute_value(defect_state, attr_name, attr_value)
            if req_event_found:
                num_match += 1
                # correct one?
                if num_match == nth_event:
                    return True, defect_state, req_attr_value
        return False, None, None

    def get_event_for_attribute_change(self, stream_defect, nth_term, attr_name, attr_value=None):
        '''Get event when specified attribute was set to given matching value'''
        logging.info('Searching for triage event n=[%d] where attribute [%s] is set to [%s]',
                     nth_term, attr_name, attr_value)

        if nth_term > 0:
            found, defect_state, value = self.seek_nth_match(stream_defect.history, nth_term, attr_name, attr_value)
        else:
            found, defect_state, value = self.seek_nth_match(reversed(stream_defect.history), abs(int(nth_term)),
                                                             attr_name, attr_value)

        return found, defect_state, value

    def get_ext_reference_id(self, stream_defect):
        '''Get external reference ID attribute value for given defect'''
        return self.get_value_for_named_attribute(stream_defect, EXT_REFERENCE_ATTR_NAME)

    def get_defect_status(self, stream_defect):
        '''Get defect status attribute value for given defect'''
        return self.get_value_for_named_attribute(stream_defect, DEFECT_STATUS_ATTR_NAME)

    def get_classification(self, stream_defect):
        '''Get classification attribute value for given defect'''
        return self.get_value_for_named_attribute(stream_defect, CLASSIFICATION_ATTR_NAME)

    def get_action(self, stream_defect):
        '''Get action attribute value for given defect'''
        return self.get_value_for_named_attribute(stream_defect, ACTION_ATTR_NAME)

    def get_defect_url(self, stream, cid):
        '''Get URL for given defect CID
        http://machine1.eng.company.com:8080/query/defects.htm?stream=StreamA&cid=1234
        '''
        return self.get_service_url('/query/defects.htm?stream=%s&cid=%s' % (stream, str(cid)))


if __name__ == '__main__':
    print("Sorry, no main here")



# -*- coding: utf-8 -*-

from setuptools import setup, find_packages
import sys

project_url = 'https://github.com/melexis/sphinx-coverity-extension'

if sys.version_info[0] == 2:
    requires = ['Sphinx>=0.6', 'docutils', 'suds', 'setuptools_scm', 'matplotlib<=3.0.3', 'urlextract-py2.7']
else:
    requires = ['Sphinx>=0.6', 'docutils', 'suds-py3', 'setuptools_scm', 'matplotlib<=3.0.3', 'urlextract']


setup(
    name='mlx.coverity',
    setup_requires=['setuptools_scm'],
    use_scm_version=True,
    url=project_url,
    license='GNU General Public License v3 (GPLv3)',
    author='Crt Mori',
    author_email='cmo@melexis.com',
    description='Sphinx coverity extension from Melexis',
    long_description=open("README.rst").read(),
    zip_safe=False,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Environment :: Web Environment',
        'Framework :: Sphinx :: Extension',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Documentation',
        'Topic :: Documentation :: Sphinx',
        'Topic :: Utilities',
    ],
    platforms='any',
    packages=find_packages(exclude=['tests', 'example']),
    include_package_data=True,
    install_requires=requires,
    namespace_packages=['mlx'],
    keywords=[
        'coverity',
        'reporting',
        'restructured text coverity report',
        'sphinx',
        'ASPICE',
        'ISO26262',
        'ASIL',
    ],
)
